@@ -14,7 +14,6 @@ require "English"
 require "erb"
 require "optparse"
 require "stringio"
-require "abbrev"
 require "highline/terminal"
 require "highline/custom_errors"
 require "highline/question"
@@ -371,10 +370,8 @@ class HighLine
   #
   # @param statement [Statement, String] what to be said
   def say(statement)
-    statement = render_statement(statement)
-    return if statement.empty?
-
-    statement = (indentation + statement)
+    statement = render_and_ident_statement(statement)
+    return statement if statement.empty?
 
     # Don't add a newline if statement ends with whitespace, OR
     # if statement ends with whitespace before a color escape code.
@@ -386,6 +383,18 @@ class HighLine
     end
   end
 
+  # Renders and indents a statement.
+  #
+  # Note: extracted here to be used by readline to render its prompt.
+  #
+  # @param statement [String] The statement to be rendered and indented.
+  # @return [String] The rendered and indented statement.
+  def render_and_ident_statement(statement)
+    statement = render_statement(statement)
+    statement = (indentation + statement) unless statement.empty?
+    statement
+  end
+
   # Renders a statement using {HighLine::Statement}
   # @param statement [String] any string
   # @return [Statement] rendered statement
@@ -538,6 +547,7 @@ class HighLine
     terminal.raw_no_echo_mode_exec do
       loop do
         character = terminal.get_character
+        raise Interrupt if character == "\u0003"
         break unless character
         break if ["\n", "\r"].include? character
 
@@ -545,6 +555,9 @@ class HighLine
         if character == "\b" || character == "\u007F"
           chopped = line.chop!
           output_erase_char if chopped && question.echo
+        elsif character == "\cU"
+          line.size.times { output_erase_char } if question.echo
+          line = ""
         elsif character == "\e"
           ignore_arrow_key
         else
