@@ -12,7 +12,6 @@ require "test_helper"
 
 require "highline"
 require "stringio"
-require "readline"
 require "tempfile"
 
 # if HighLine::CHARACTER_MODE == "Win32API"
@@ -349,92 +348,42 @@ class TestHighLine < Minitest::Test
     assert_equal(4, @output.string.count("\b"))
   end
 
-  def test_readline_mode
-    #
-    # Rubinius (and JRuby) seems to be ignoring
-    # Readline input and output assignments. This
-    # ruins testing.
-    #
-    # But it doesn't mean readline is not working
-    # properly on rubinius or jruby.
-    #
-
-    terminal = @terminal.terminal
-
-    if terminal.jruby? || terminal.rubinius? || terminal.windows?
-      skip "We can't test Readline on JRuby, Rubinius and Windows yet"
-    end
-
-    # Creating Tempfiles here because Readline.input
-    #   and Readline.output only accepts a File object
-    #   as argument (not any duck type as StringIO)
-
-    temp_stdin  = Tempfile.new "temp_stdin"
-    temp_stdout = Tempfile.new "temp_stdout"
-
-    Readline.input  = @input  = File.open(temp_stdin.path, "w+")
-    Readline.output = @output = File.open(temp_stdout.path, "w+")
-
-    @terminal = HighLine.new(@input, @output)
-
-    @input << "any input\n"
+  def test_erase_line_does_not_enter_prompt
+    @input << "\cU\cU\cU\cU\cU\cU\cU\cU\cU"
     @input.rewind
-
-    answer = @terminal.ask("Prompt:  ") do |q|
-      q.readline = true
-    end
-
-    @output.rewind
-    output = @output.read
-
-    assert_equal "any input", answer
-    assert_match "Prompt:  any input\n", output
-
-    @input.close
-    @output.close
-    Readline.input  = STDIN
-    Readline.output = STDOUT
-  end
-
-  def test_readline_mode_with_limit_set
-    temp_stdin  = Tempfile.new "temp_stdin"
-    temp_stdout = Tempfile.new "temp_stdout"
-
-    Readline.input  = @input  = File.open(temp_stdin.path, "w+")
-    Readline.output = @output = File.open(temp_stdout.path, "w+")
-
-    @terminal = HighLine.new(@input, @output)
-
-    @input << "any input\n"
-    @input.rewind
-
-    answer = @terminal.ask("Prompt:  ") do |q|
-      q.limit = 50
-      q.readline = true
-    end
-
-    @output.rewind
-    output = @output.read
-
-    assert_equal "any input", answer
-    assert_equal "Prompt:  any input\n", output
-
-    @input.close
-    @output.close
-    Readline.input  = STDIN
-    Readline.output = STDOUT
-  end
-
-  def test_readline_on_non_echo_question_has_prompt
-    @input << "you can't see me"
-    @input.rewind
-    answer = @terminal.ask("Please enter some hidden text: ") do |q|
-      q.readline = true
+    answer = @terminal.ask("Please enter your password: ") do |q|
       q.echo = "*"
     end
-    assert_equal("you can't see me", answer)
-    assert_equal("Please enter some hidden text: ****************\n",
-                 @output.string)
+    assert_equal("", answer)
+    assert_equal("Please enter your password: \n", @output.string)
+
+    # There's no backspaces as the line is already empty
+    assert_equal(0, @output.string.count("\b"))
+  end
+
+  def test_after_some_chars_erase_line_does_not_enter_prompt_when_ascii
+    @input << "apple\cU\cU\cU\cU\cU\cU\cU\cU"
+    @input.rewind
+    answer = @terminal.ask("Please enter your password: ") do |q|
+      q.echo = "*"
+    end
+    assert_equal("", answer)
+
+    # There's only enough backspaces to clear the given string
+    assert_equal(5, @output.string.count("\b"))
+  end
+
+
+  def test_after_some_chars_erase_line_does_not_enter_prompt_when_utf8
+    @input << "maçã\cU\cU\cU\cU\cU\cU\cU\cU"
+    @input.rewind
+    answer = @terminal.ask("Please enter your password: ") do |q|
+      q.echo = "*"
+    end
+    assert_equal("", answer)
+
+    # There's only enough backspaces to clear the given string
+    assert_equal(4, @output.string.count("\b"))
   end
 
   def test_character_reading
@@ -817,19 +766,49 @@ class TestHighLine < Minitest::Test
   end
 
   def test_default_with_non_String_objects
-    # With a non-string object, it should not show
+    # With a non-string object, it should try to coerce it to String.
+    # If the coercion is not possible, do not show
     #   any 'default' at prompt line. And should
     #   return the "default" object, without conversion.
 
     @input << "\n"
     @input.rewind
 
+    default_integer_object = 42
+
+    answer = @terminal.ask("Question:  ") do |q|
+      q.default = default_integer_object
+    end
+
+    assert_equal default_integer_object, answer
+    assert_equal "Question:  |42|  ", @output.string
+
+    @input.truncate(@input.rewind)
+    @input << "\n"
+    @input.rewind
+    @output.truncate(@output.rewind)
+
     default_non_string_object = Object.new
 
     answer = @terminal.ask("Question:  ") do |q|
       q.default = default_non_string_object
     end
 
+    assert_equal default_non_string_object, answer
+    assert_equal "Question:  |#{default_non_string_object}|  ", @output.string
+
+    @input.truncate(@input.rewind)
+    @input << "\n"
+    @input.rewind
+    @output.truncate(@output.rewind)
+
+    default_non_string_object = Object.new
+
+    answer = @terminal.ask("Question:  ") do |q|
+      q.default = default_non_string_object
+      q.default_hint_show = false
+    end
+
     assert_equal default_non_string_object, answer
     assert_equal "Question:  ", @output.string
   end
@@ -1270,6 +1249,20 @@ class TestHighLine < Minitest::Test
     assert_equal("maçã", answer)
   end
 
+  def test_echo_false_with_ctrl_c_interrupts
+    @input << "String with a ctrl-c at the end \u0003 \n"
+    @input.rewind
+    @answer = nil
+
+    assert_raises(Interrupt) do
+      @answer = @terminal.ask("Type:  ") do |q|
+        q.echo = false
+      end
+    end
+
+    assert_nil @answer
+  end
+
   def test_range_requirements
     @input << "112\n-541\n28\n"
     @input.rewind
@@ -1358,6 +1351,25 @@ class TestHighLine < Minitest::Test
                   "?  ", @output.string)
   end
 
+  def test_range_requirements_with_array_of_strings
+    @input.truncate(@input.rewind)
+    @input << "z\nx\nb\n"
+    @input.rewind
+    @output.truncate(@output.rewind)
+
+    answer = @terminal.ask("Letter a, b, or c? ") do |q|
+      q.in = %w[ a b c ]
+    end
+    assert_equal("b", answer)
+    assert_equal("Letter a, b, or c? " \
+                  "Your answer isn't within the expected range " \
+                  "(included in [\"a\", \"b\", \"c\"]).\n" \
+                  "?  " \
+                  "Your answer isn't within the expected range " \
+                  "(included in [\"a\", \"b\", \"c\"]).\n" \
+                  "?  ", @output.string)
+  end
+
   def test_reask
     number = 61_676
     @input << "Junk!\n" << number << "\n"
@@ -1577,6 +1589,55 @@ class TestHighLine < Minitest::Test
     assert_equal("Gray, James Edward", answer)
   end
 
+  class ZeroToTwentyFourValidator
+    def self.valid?(answer)
+      (0..24).include? answer.to_i
+    end
+
+    def self.inspect
+      "(0..24) rule"
+    end
+  end
+
+  def test_validation_with_custom_validator_class
+    @input << "26\n25\n24\n"
+    @input.rewind
+
+    answer = @terminal.ask("What hour of the day is it?:  ", Integer) do |q|
+      q.validate = ZeroToTwentyFourValidator
+    end
+
+    assert_equal(24, answer)
+    assert_equal("What hour of the day is it?:  " \
+                "Your answer isn't valid (must match (0..24) rule).\n" \
+                "?  Your answer isn't valid (must match (0..24) rule).\n" \
+                "?  ", @output.string)
+  end
+
+  require 'dry/types'
+
+  module Types
+    include Dry.Types
+  end
+
+  def test_validation_with_dry_types
+    @input << "random string\nanother uncoercible string\n42\n"
+    @input.rewind
+
+    answer = @terminal.ask("Type a number:  ", Integer) do |q|
+      q.validate = Types::Coercible::Integer
+    end
+
+    assert_equal(42, answer)
+    assert_match(Regexp.new(<<~REGEXP.chomp),
+      Type a number:  Your answer isn't valid .must match .*Dry.*Types.*Integer.*..
+      \\\?  Your answer isn't valid .must match .*Dry.*Types.*Integer.*..
+      \\\?
+    REGEXP
+      @output.string
+    )
+  end
+
   def test_validation_with_overriding_static_message
     @input << "Not valid answer\n" \
               "42\n"
